; ---------------------------------------------------------------------------
; Fire effect by Jestyr aka Sam Izzo (taran@alphalink.com.au)
;
; Written on: 27-12-1997
;
; Updated on: 08-01-1998
;     - rearranged palette setting to save 1 byte!  instead of doing all in
;       one, did one loop for each of red, green, blue, and one loop for
;       last 64 colours of bright white
;     - used rep movs dword ptr es:[si], fs:[di]
;
; Updated on: 13-01-1998
;     - organise code so no swapping of segment registers was necessary
;       ie segment registers are set at start only, saved 1 byte!!
;     - some cosmetic changes to the code
;
; Updated on: 21-01-1999 (yes, I came back to it a year later :))
;     - saved 1 byte by doing rep movs word ptr instead of dword ptr
;
; Updated on: 17-07-1999 (hm.. this has been going awhile :))
;     - 3 bytes smaller!  Replaced mov instructions where ax was used with
;       another register (only bx/cx/dx) to an xchg instruction
;
; Updated on: 24-07-1999
;     - oh my god!  22 bytes smaller!  removed all memory allocation and made
;       use of the fact that .com files are initially allocated all available
;       memory. however, if .com is loaded at f000h or higher then some other
;       part of memory will be overwritten, possibly causing a crash.  could
;       put a simple check for this but I never had checks for free memory,
;       so why bother?
;     - changed escape check (had dec rb, changed to dec rw) and mode change
;
; Updated on: 15-05-2000
;     - modified buffer clearing code to loop instead of duplicating code for
;       the two buffers (saved 1 byte)
;     - replaced dec/jnz in hotspot generating code with loop instruction
;
; Updated on: 12-10-2000
;     - modified buffer clear routine to use bx instead of byte ptr [count+1]
;       as the counter.
;
; Updated on: 08-11-2000
;     - now loads gs with 0a000h from the psp (byte 2/3 contains a000)
;     - made obvious optimisation on palette setting code that i had missed
;       before.  instead of looping 63 times to do three out dx, al instructions
;       it now loops 63*3 times to do one out dx, al instruction (saved 2 bytes)
;     - loaded al directly for the first pixel in the averaging loop (saved
;       1 byte)
;
; Updated on: 26/01/2001
;     - no longer loads gs from the psp, as bytes 2/3 are different under win2k
;     - buffer clearing code changed.  ch is loaded with 7eh to make cx contain
;       between 32256d and 32511d (depending on initial value of cl) -- this is
;       ok as the buffer is actually 65535 bytes in size.  previously i was only
;       clearing 64000 bytes.
;     - removed cmp/jmp that was used to not draw the last 4 lines to the
;       screen.  hotspots are drawn lower down now.  this saved 6 bytes!
;     - number of hotspots is now set via ch and not cx (saved 1 byte)
;     - minor cosmetic changes to code
;     - modified this version to compile with NASM rather than A86
;
; This is the NASM version, my current assembler of choice.
;
; It's fast enough on my P100.  I was aiming to get it fast, decent looking,
; and small.  I hope I managed alright.  It's guaranteed to be slow on a 386,
; it's OK on a 486, and it should be fine on any kind of Pentium.  If you can
; get it smaller, mail me please!
;
; It doesn't check for free memory, so make sure you have about 130k free.
; It also doesn't check for an 80386 or VGA.
;
; It should be 173 bytes with NASM v0.98
;
; To compile:
;
;   nasm fire.asm -o fire.asm
;
; Escape quits.
;
; ---------------------------------------------------------------------------

      ; first buffer is at end of code (roughly)
      mov dx, cs ; calculate segment for end of code (ie offset zero)
      add dx, byte 20h ; code+psp is approx. 512 bytes (32 pages)

      ; second buffer is 64k after first buffer
      mov ax, dx ; second buffer segment
      add ax, 1000h ; 4k pages (hope this doesn't overflow)

      ; second buffer is accessed through fs
      mov fs, ax

      ; zero the buffers
      mov es, ax
zero:
      xor di, di
      xor ax, ax
      mov ch, 7eh ; 32256d <= cx <= 32511d (depending on cl)
      rep stosw

      push dx
      pop es

      dec bx
      jp short zero

      push word 0a000h
      pop gs

      push es
      pop ds

      ; es -> buf1
      ; fs -> buf2
      ; ds -> es -> buf1
      ; gs -> 0a000h

      ; set vid mode to 320x200x256
      mov al, 13h ; ah is already 0
      int 10h

      ; here's some pseudocode for how i set the palette:
      ; for i = 0 to 63
      ; begin
      ;    setpal(i, i, 0, 0)
      ;    setpal(i+64, 63, i, 0)
      ;    setpal(i+128, 63, 63, i)
      ;    setpal(i+192, 63, 63, 63)
      ; end

      ; set up the palette
      mov dx, 3c8h
      xor al, al
      out dx, al

      inc dx

      mov cl, 64 ; ch is already 0

red_loop:

      mov al, cl
      neg al
      out dx, al
      xor al, al
      out dx, al
      out dx, al

      loop red_loop

      mov cl, 64

green_loop:

      mov al, 63
      out dx, al
      mov al, cl
      neg al
      out dx, al
      xor al, al
      out dx, al

      loop green_loop

      mov cl, 64

blue_loop:

      mov al, 63
      out dx, al
      out dx, al
      mov al, cl
      neg al
      out dx, al

      loop blue_loop

      mov cl, 64*3-1
      mov al, cl

white_loop:

      out dx, al

      loop white_loop

      mov dl, 0ffh ; dl is the hotspot colour

      ; es -> buf1
      ; fs -> buf2
      ; ds -> es -> buf1
      ; gs -> 0a000h

      ; main loop - generate hotspots, average, etc

mainloop:

      mov ch, 2 ; change this to draw more hotspots, around 512 looks good
                ; under windows.  mov cx, 520 can be used but this is bigger

      ; draw hotspots at random positions on the bottom row of buf1 (ds)

  randloop:

      in al, 40h ; get a value from the timer to use as a random number
      add ax, bp ; bp is the seed
      add bp, ax ; fiddle with it

      ; divide it a bit so it doesn't go over 512
      xchg bx, ax ; put it in bx
      shr bx, 7

      ; hotspot on buf1 (ds)
      mov [bx+64320], dl ; 201*320

      loop randloop

      ; segment overrides take more space and execution time, so ds points
      ; to the buffer which gets accessed the most in the averaging loop

  averageloop:

      ; add up our four pixels..
      ; i use buf[x, y]+buf[x, y-1]+buf[x-1, y]+buf[x+1, y]
      ; this looks quite good, but you can do whatever you want, really

      ; average from buf1 (ds)
      mov al, [di]
      mov bl, [di-320]
      add ax, bx
      mov bl, [di-1]
      add ax, bx
      mov bl, [di+1]
      add ax, bx

      sar ax, 2 ; ..divide by four..

      ; ..and al contains the averaged value, which we decay a bit
      cmp al, 2
      jle short nodecay
      sub al, 2

  nodecay:

      mov [fs:di-320], al ; draw to buf2 (fs)
      mov [gs:di], al ; draw to video memory

      dec di

      cmp di, 30000 ; no need to average the entire screen if the fire never
      ja short averageloop ; gets that high

      ; blit buffer at fs to buffer at es (copy buf2 to buf1)
      mov cx, 17320 ; no need to flip the entire buffer
      mov si, di ; when we come out of the averaging loop, di = 30000d
      rep fs movsw ; source segment override move

;;; uncomment the following to slow the effect (this waits for vertical retrace)
;      mov dx, 3dah
;    retrace:
;      in al, dx
;      and al, 08h
;      jz short retrace
;      mov dl, 0ffh

      ; check for escape (scancode of 1)
      in al, 60h
      dec ax ; ah seems to be zero most of the time
      jnz short mainloop

      ; back to text mode
      mov al, 3 ; ah should be zero if escape was pressed
      int 10h

      ret ; cheeky way to quit

; ..and the buffers appear down here somewhere in RAM

; - eof fire.asm ------------------------------------------------------------

